@@ -1,7 +1,4 @@
 """
-Steps
------
-
 Common and reusable steps
 """
 from .common import NodeTagger, NodeTagCopy, TextParser, RollupTransformer, TagsToKeyValuePairExtractor, \
